//provides a proxy which connects to a cam daemon socket on behalf of pmt files.

#include "pmt-proxy.h"
#include "common.h"

int pmt_proxy::dir_fd = -1;
char pmt_proxy::pmt_file_path[32] = {0};
char pmt_proxy::pmt_filename_pattern[32] = {0};
char pmt_proxy::camd_socket_path[32] = {0};
bool pmt_proxy::ready = false;
pmt_translation_t pmt_proxy::translation = NO_TRANSLATION;

struct pmt_proxy_connection pmt_proxy::connections[MAX_CONNECTION_COUNT] = {
	{0, -1}, /* pmt0 */
	{0, -1}  /* pmt1 */
};

pmt_proxy::pmt_proxy( void ) {}
pmt_proxy::~pmt_proxy() {}

/*static */ void pmt_proxy::start( pmt_translation_t device, char *camd_socket )
{
	//--- start the proxy
	//set pmt path and file name pattern
	translation = device;
	strcpy(camd_socket_path, camd_socket);
	
	switch( translation )
	{
		case DGSTATION_DEVICES:
			strcpy(pmt_file_path, "/tmp/");
			strcpy(pmt_filename_pattern, "pmt%d.tmp");
			break;
		
		default:
			_slog(NO_TRANS, _DEBUG_BASIC, 1, "[PMT Proxy] Invalid PMT type selected. PMT proxy not started.", 2);
			return;
	}
	
	//configure and set the signal handler
	struct sigaction signal_action;							// action descriptor struct
	
	signal_action.sa_handler = pmt_proxy::event_handler;	// define the event handler for the signal
	sigemptyset(&signal_action.sa_mask);					// no other signals will be masked during the call to 
															// the event handler
	signal_action.sa_flags = 0;
	sigaction(SIGRTMIN + 1, &signal_action, NULL);			// bind the action to the signal descriptor
	
	//open the pmt file directory and set it to trigger our signal (SIGRTMIN) on directory events.
	dir_fd = open(pmt_file_path, O_RDONLY);
	if( dir_fd >= 0 )
	{
		fcntl(dir_fd, F_SETSIG, SIGRTMIN + 1);
		fcntl(dir_fd, F_NOTIFY, DN_MODIFY | DN_CREATE | DN_DELETE | DN_MULTISHOT);
	}
	else
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 2, _logmf("[PMT Proxy] Failed to open %s directory. PMT proxy not started.", pmt_file_path), 1);
		return;
	}
	
	ready = true;
	_slog(NO_TRANS, _DEBUG_BASIC, 1, "[PMT Proxy] Started successfully.", 1);
	
	//call the handler at start-up
	event_handler(SIGRTMIN + 1);
}

/* static */ void pmt_proxy::stop( void )
{
	//--- stop the proxy
	fcntl(dir_fd, F_NOTIFY, 0);
	close( dir_fd );
	ready = false;
}

/* static */ void pmt_proxy::event_handler(int signal)
{
	char pmt_file_name[32];
	char pmt_full_path[64];
	int pmt_fd;
	struct stat pmt_info;
	unsigned char *pmt_original_buffer;

	unsigned char pmt_translated_buffer[256];
	unsigned int pmt_translated_buffer_length;

	int camd_socket_fd;
	
	//this 100ms delay is necessary for newly created pmt files to hatch,
	//otherwise new pmt files are reported as 0-bytes.
	usleep(100000); 
	
	for(int i=0;i<MAX_CONNECTION_COUNT;i++)
	{
		sprintf(pmt_file_name, pmt_filename_pattern, i);
		sprintf(pmt_full_path, "%s%s", pmt_file_path, pmt_file_name);
		if( (pmt_fd = open(pmt_full_path, O_RDONLY)) >= 0 )
		{
			if( fstat(pmt_fd, &pmt_info) == 0 )
			{
				if( pmt_info.st_mtime != connections[i].timestamp )
				{
					//close connection if necessary)
					if( connections[i].camd_socket_fd >= 0 )
					{
						close( connections[i].camd_socket_fd );
						connections[i].camd_socket_fd = -1;
						usleep(100000);	//sleep for 100ms between disconnect/connect
										//for safety and to play nice with the cam.
					}
					
					//pmt changed and file is not empty, load pmt file
					if( pmt_info.st_size > 0 )
					{
						if( (pmt_original_buffer = load(pmt_fd, (size_t)pmt_info.st_size)) )
						{
							connections[i].timestamp = pmt_info.st_mtime;
							
							//file loaded successfully, translate to Enigma 1 format
							_flog(NO_TRANS, _DEBUG_EXTENDED, 1, "[PMT Proxy] Raw File PMT Data:", 2, pmt_original_buffer, (unsigned int)pmt_info.st_size, 16, 1, 2);
							translate(i, pmt_translated_buffer, &pmt_translated_buffer_length, 
								pmt_original_buffer, (unsigned int)pmt_info.st_size);
							free( pmt_original_buffer );
							
							// ---------------------------- connect to cam socket -------------------------------
							if( (camd_socket_fd = connect_unix_socket( camd_socket_path )) >= 0 )
							{
								connections[i].camd_socket_fd = camd_socket_fd;
								
								//write pmt data, and leave connection open.
								if( write(camd_socket_fd, pmt_translated_buffer, pmt_translated_buffer_length) != pmt_translated_buffer_length )
									_slog(NO_TRANS, _DEBUG_BASIC, 2, _logmf("[PMT Proxy] Failed to write full PMT data to cam socket %s.", camd_socket_path), 1);
							}
							else 
								_slog(NO_TRANS, _DEBUG_BASIC, 2, _logmf("[PMT Proxy] Failed connect to cam socket %s (%d).", camd_socket_path), 1);
						}
					}
				}
			}
			else _slog(NO_TRANS, _DEBUG_BASIC, 2, _logmf("[PMT Proxy] Failed to get PMT file info for %s.", pmt_file_path), 1);
			close( pmt_fd );
		}
		else
		{
			//---------------------- pmt not found ------------------------
			if( connections[i].timestamp > 0 )
			{
				//pmt deleted, close cam socket connection if necessary.
				if( connections[i].camd_socket_fd > 0 )
				{
					close( connections[i].camd_socket_fd );
					connections[i].camd_socket_fd = -1;
				}
				connections[i].timestamp = 0;
			}		
			//-------------------------------------------------------------
		}
	}
}

/* static */ unsigned char *pmt_proxy::translate(unsigned int pmt_index, unsigned char *dest, unsigned int *dest_length, unsigned char *src, unsigned int src_length)
{
	switch( translation )
	{
		case NO_TRANSLATION:
		{
			memcpy(dest, src, src_length);
			*dest_length = src_length;
			break;
		}

		case DGSTATION_DEVICES:
		{
			//create pmt header
			memcpy(dest, "\x9F\x80\x32\x00\x00\x00\x00\xFF\xFF\x00\x00\x13\x00\x81\x08\x00\x00\x00\x00\x00\x00\xFF\xFF\x82\x02\x01\x01\x84\x02\x00\x00", 31);
			
			//set onid in header.
			switch( src[1] )
			{
				case 0x80:
					dest[21] = 0x10; 
					break;
				
				case 0xB0:
					dest[21] = 0x01;
					break;
				
				default:
					dest[21] = 0x10; 
			}
			dest[22] = 0x00;
			
			//set sid in header.
			dest[7] = src[3];
			dest[8] = src[4]; 
			
			//set descramble index according to pmt index
			//(descramble mask 0x01 for pmt0, descramble mask 0x02 for pmt1)
			dest[25] = pmt_index + 1;
			
			//copy pmt body
			memcpy(dest + 31, src + 7, src_length - 7 - 4);
			
			//calculate final translated pmt length
			*dest_length = 31 + src_length - 7 - 4;
			break;
		}
	}
	return dest;
}

/* static */ unsigned char *pmt_proxy::load(int fd, size_t size)
{
	unsigned char *buffer;
	
	if( (buffer = (unsigned char *)malloc(size)) == NULL )
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 2, "[PMT Proxy] Could not allocate memory for PMT file loading.", 1);
		return NULL;
	}
	
	if( read(fd, buffer, size) != size )
	{
		free( buffer );
		_slog(NO_TRANS, _DEBUG_BASIC, 2, "[PMT Proxy] Failed to load complete PMT file to memory.", 1);
		return NULL;
	}

	return buffer;
}

/* static */ int pmt_proxy::connect_unix_socket(char *path)
{
	int fd;
	struct sockaddr_un socket_address;
	socket_address.sun_family = AF_UNIX;
	strcpy(socket_address.sun_path, path);
	
	if( (fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 ) return -1;
	
	if( connect(fd, (struct sockaddr *)&socket_address, sizeof(socket_address)) < 0 )
	{
		close( fd );
		return -1;
	}
	else return fd;
}
